from unittest.mock import patch, MagicMock
from django.test import override_settings
from rest_framework.reverse import reverse
from rest_framework import status
from arkindex_common.ml_tool import MLToolType
from arkindex.dataimport.models import DataImport
from arkindex.documents.models import Element, Corpus
from arkindex.project.tests import FixtureAPITestCase


@patch('arkindex_common.ml_tool.MLTool.get')
class TestWorkflows(FixtureAPITestCase):
    """
    Test workflows from Arkindex existing elements
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume = Element.objects.get(name='Volume 1')
        cls.pages = Element.objects.get_descending(cls.volume.id).filter(
            type__folder=False,
            type__hidden=False,
            zone__isnull=False,
        )
        cls.private_corpus = Corpus.objects.create(name='private')
        cls.private_volume = cls.private_corpus.elements.create(
            type=cls.private_corpus.types.create(slug='volume', folder=True),
            name='Hidden',
        )
        cls.ml_tools_sample = [{'type': 'dla', 'slug': 'line-detector', 'parents': []}]

    @override_settings(
        ARKINDEX_TASKS_IMAGE='tasks:latest',
        PONOS_RECIPE={'env': {'SOME_VAR': 'somevalue'}}
    )
    def test_elements_workflow(self, ml_get_mock):
        dla_mock, reco_mock = MagicMock(), MagicMock()
        dla_mock.type, dla_mock.slug = MLToolType.DLAnalyser, 'line-detector'
        reco_mock.type, reco_mock.slug = MLToolType.Recognizer, 'text-recognizer'
        dla_mock.get_task_recipe.return_value = {'image': 'im1', 'command': 'cmd1'}
        reco_mock.get_task_recipe.return_value = {'image': 'im2', 'command': 'cmd2'}
        tools = {dla_mock.slug: dla_mock, reco_mock.slug: reco_mock}
        ml_get_mock.side_effect = lambda directory, type, slug: tools[slug]

        self.client.force_login(self.user)
        ml_tools = [
            {'type': 'dla', 'slug': 'line-detector', 'parents': []},
            {
                'type': 'recognizer',
                'slug': 'text-recognizer',
                'parents': [{'type': 'dla', 'slug': 'line-detector'}],
            }
        ]

        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': ml_tools,
                'corpus': str(self.corpus.id)
            },
            format='json',
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        workflow = dataimport.workflow

        self.assertDictEqual(data, {
            'id': str(dataimport.id),
            'state': 'unscheduled',
            'corpus': str(self.volume.corpus.id),
            'files': [],
            'mode': 'elements',
            'payload': {
                'corpus_id': str(self.volume.corpus.id),
                'ml_tools': [
                    {
                        'slug': 'line-detector',
                        'type': 'dla',
                        'parents': [],
                    }, {
                        'slug': 'text-recognizer',
                        'type': 'recognizer',
                        'parents': [{'slug': 'line-detector', 'type': 'dla'}]
                    }
                ],
                'chunks': 1
            },
            'revision': dataimport.revision,
            'workflow': reverse(
                'ponos:workflow-details',
                kwargs={'pk': workflow.id},
                request=response.wsgi_request
            )
        })
        self.assertEqual(workflow.recipes['initialisation'].image, 'tasks:latest')
        self.assertEqual(
            workflow.recipes['initialisation'].command,
            'python -m arkindex_tasks.init_elements --corpus-id {} --chunks-number 1'
            .format(str(self.corpus.id))
        )

    def test_volume_workflow(self, ml_get_mock):
        page = self.pages[0]
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id),
                'element': str(self.volume.id)
            },
            format='json'
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        workflow = dataimport.workflow

        self.assertDictEqual(data, {
            'id': str(dataimport.id),
            'state': 'unscheduled',
            'mode': 'elements',
            'corpus': str(page.corpus.id),
            'payload': {
                'corpus_id': str(page.corpus.id),
                'chunks': 1,
                'element': str(self.volume.id),
                'ml_tools': [],
                'folder_id': str(self.volume.id),
            },
            'files': [],
            'revision': dataimport.revision,
            'workflow': reverse(
                'ponos:workflow-details',
                kwargs={'pk': workflow.id},
                request=response.wsgi_request
            )
        })

    def test_single_page_workflow(self, ml_get_mock):
        page = self.pages[0]
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id),
                'element': str(self.pages.first().id)
            },
            format='json'
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertDictEqual(data.get('payload'), {
            'corpus_id': str(page.corpus.id),
            'chunks': 1,
            'element': str(self.pages.first().id),
            'folder_id': str(self.volume.id),
            'ml_tools': []
        })

    @override_settings(
        ARKINDEX_TASKS_IMAGE='tasks:latest',
    )
    def test_chunks(self, ml_get_mock):
        dla_mock, reco_mock = MagicMock(), MagicMock()
        dla_mock.type, dla_mock.slug = MLToolType.DLAnalyser, 'line-detector'
        reco_mock.type, reco_mock.slug = MLToolType.Recognizer, 'text-recognizer'
        dla_mock.get_task_recipe.return_value = {'image': 'im1', 'command': 'cmd1'}
        reco_mock.get_task_recipe.return_value = {'image': 'im2', 'command': 'cmd2'}
        tools = {dla_mock.slug: dla_mock, reco_mock.slug: reco_mock}
        ml_get_mock.side_effect = lambda directory, type, slug: tools[slug]
        ml_tools = [
            {
                'slug': 'line-detector',
                'type': 'dla',
                'parents': [],
            }, {
                'slug': 'text-recognizer',
                'type': 'recognizer',
                'parents': [{'slug': 'line-detector', 'type': 'dla'}]
            }
        ]
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': ml_tools,
                'element': str(self.volume.id),
                'corpus': str(self.corpus.id),
                'chunks': 3,
                'thumbnails': True
            },
            format='json'
        )

        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        workflow = dataimport.workflow

        self.assertDictEqual(
            dataimport.payload,
            {
                'chunks': 3,
                'thumbnails': True,
                'element': str(self.volume.id),
                'corpus_id': str(self.corpus.id),
                'ml_tools': ml_tools,
                'folder_id': str(self.volume.id)
            }
        )
        # 1 import, 3 thumbnails tasks, 3 times 2 machine learning tasks
        self.assertEqual(len(workflow.recipes), 10)
        self.assertListEqual(
            [(name, r.command, r.parents) for name, r in workflow.recipes.items()],
            [
                (
                    'initialisation',
                    'python -m arkindex_tasks.init_elements --corpus-id {} --chunks-number 3 --element-id {}'
                    .format(self.corpus.id, self.volume.id),
                    []
                ),
                ('ml_line-detector_1', 'cmd1', ['initialisation']),
                ('ml_line-detector_2', 'cmd1', ['initialisation']),
                ('ml_line-detector_3', 'cmd1', ['initialisation']),
                ('ml_text-recognizer_1', 'cmd2', ['ml_line-detector_1']),
                ('ml_text-recognizer_2', 'cmd2', ['ml_line-detector_2']),
                ('ml_text-recognizer_3', 'cmd2', ['ml_line-detector_3']),
                (
                    'thumbnails_1',
                    'python3 -m arkindex_tasks.generate_thumbnails /data/initialisation/elements_chunk_1.json',
                    ['initialisation']
                ), (
                    'thumbnails_2',
                    'python3 -m arkindex_tasks.generate_thumbnails /data/initialisation/elements_chunk_2.json',
                    ['initialisation']
                ), (
                    'thumbnails_3',
                    'python3 -m arkindex_tasks.generate_thumbnails /data/initialisation/elements_chunk_3.json',
                    ['initialisation']
                )
            ]
        )

    def test_element_workflow_requires_login(self, ml_get_mock):
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'element': str(self.volume.id)
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_element_workflow_acl(self, ml_get_mock):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': self.ml_tools_sample,
                'element': str(self.private_volume.id)
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('element', data)
        self.assertListEqual(
            data['element'],
            ['Invalid pk "{}" - object does not exist.'.format(self.private_volume.id)]
        )

    def test_tools_empty(self, ml_get_mock):
        self.client.force_login(self.user)
        dla_mock = MagicMock()
        ml_get_mock.side_effect = dla_mock
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id)
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertListEqual(dataimport.payload['ml_tools'], [])

    def test_elements_name_filter(self, ml_get_mock):
        self.client.force_login(self.user)
        dla_mock = MagicMock()
        ml_get_mock.side_effect = dla_mock
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id),
                'name': self.pages.first().name[2:5]
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertDictEqual(dataimport.payload, {
            'chunks': 1,
            'corpus_id': str(self.corpus.id),
            'ml_tools': [],
            'name_contains': self.pages.first().name[2:5],
        })

    def test_elements_type_filter(self, ml_get_mock):
        self.client.force_login(self.user)
        dla_mock = MagicMock()
        ml_get_mock.side_effect = dla_mock
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id),
                'type': 'page'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        self.assertDictEqual(dataimport.payload, {
            'chunks': 1,
            'corpus_id': str(self.corpus.id),
            'elt_type': 'page',
            'ml_tools': []
        })

    def test_elements_empty_queryset(self, ml_get_mock):
        self.client.force_login(self.user)
        dla_mock = MagicMock()
        ml_get_mock.side_effect = dla_mock
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id),
                'type': 'volume',
                'name': 'ahahah'
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('__all__', data)
        self.assertListEqual(
            data['__all__'],
            ['No element match filtering']
        )

    def test_elements_empty_selection(self, ml_get_mock):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id),
                'selection': True
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('__all__', data)
        self.assertListEqual(
            data['__all__'],
            ['No element match filtering']
        )

    def test_wrong_type(self, ml_get_mock):
        self.client.force_login(self.user)
        dla_mock = MagicMock()
        ml_get_mock.side_effect = dla_mock
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id),
                'type': 'arachnid',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('type', data)
        self.assertListEqual(
            data['type'],
            ['Unrecognized element type filter']
        )

    def test_mixed_element_filters(self, ml_get_mock):
        self.client.force_login(self.user)
        dla_mock = MagicMock()
        ml_get_mock.side_effect = dla_mock
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id),
                'element': str(self.pages.first().id),
                'type': 'page',
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('__all__', data)
        self.assertListEqual(
            data['__all__'],
            ['Filtering parameters (name, type) cannot be used with element parameter']
        )

    def test_mixed_selection_filters(self, ml_get_mock):
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id),
                'element': str(self.pages.first().id),
                'selection': True
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        data = response.json()
        self.assertIn('__all__', data)
        self.assertListEqual(
            data['__all__'],
            ['Filtering parameters (element, name, type) cannot be used with selection parameter']
        )

    def test_elements_name_escape(self, ml_get_mock):
        self.client.force_login(self.user)
        dla_mock = MagicMock()
        ml_get_mock.side_effect = dla_mock
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id),
                'name': self.pages.first().name
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        workflow = dataimport.workflow
        self.assertEqual(
            workflow.recipes['initialisation'].command,
            'python -m arkindex_tasks.init_elements --corpus-id {} --chunks-number 1 --name-contains \'{}\''
            .format(str(self.corpus.id), str(self.pages.first().name))
        )

    def test_elements_selection_filter(self, ml_get_mock):
        self.client.force_login(self.user)
        self.client.post(
            reverse('api:elements-selection'),
            data={'ids': [self.pages.first().id]},
            format='json'
        )
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id),
                'selection': True
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        workflow = dataimport.workflow
        self.assertDictEqual(dataimport.payload, {
            'chunks': 1,
            'corpus_id': str(self.corpus.id),
            'ml_tools': [],
            'elements': [str(self.pages.first().id)]
        })
        self.assertEqual(
            workflow.recipes['initialisation'].command,
            'python -m arkindex_tasks.init_elements --corpus-id {} --chunks-number 1 --dataimport-id {}'
            .format(str(self.corpus.id), str(dataimport.id))
        )

    def test_thumbnails_generation_only(self, ml_get_mock):
        """
        Generating thumbnails without any workflow must generate an import task
        which do not to retrieve filtered elements children
        """
        self.client.force_login(self.user)
        corpus_type = self.corpus.types.first()
        response = self.client.post(
            reverse('api:corpus-workflow'),
            {
                'ml_tools': [],
                'corpus': str(self.corpus.id),
                'chunks': 3,
                'type': corpus_type.slug,
                'thumbnails': True
            },
            format='json'
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data['id'])
        workflow = dataimport.workflow
        self.assertEqual(
            workflow.recipes['initialisation'].command,
            'python -m arkindex_tasks.init_elements --corpus-id {} --chunks-number 3 --type {} --skip-children'
            .format(str(self.corpus.id), corpus_type.slug)
        )
