from django.utils.decorators import method_decorator
from django.views.decorators.cache import cache_page
from rest_framework.generics import RetrieveAPIView
from rest_framework.exceptions import PermissionDenied
from arkindex_common.enums import TranscriptionType
from arkindex.documents.models import Element, Corpus
from arkindex.documents.serializers.search import IIIFSearchQuerySerializer
from arkindex.documents.serializers.iiif import (
    FolderManifestSerializer,
    ElementAnnotationListSerializer,
    TranscriptionSearchAnnotationListSerializer,
)
from arkindex.documents.search import search_transcriptions_filter_post
from arkindex.project.elastic import ESTranscription
from arkindex.project.mixins import SearchAPIMixin


class FolderManifest(RetrieveAPIView):
    """
    Get a IIIF manifest for a specific volume
    """

    serializer_class = FolderManifestSerializer
    openapi_overrides = {
        'operationId': 'RetrieveFolderManifest',
        'description': 'Retrieve an IIIF manifest for a folder element.',
        'security': [],
        'tags': ['iiif'],
    }

    def get_queryset(self):
        return Element.objects.filter(
            corpus__in=Corpus.objects.readable(self.request.user),
            type__folder=True,
        )

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class ElementAnnotationList(RetrieveAPIView):
    """
    Get a IIIF annotation list for transcriptions of a specific page
    """

    serializer_class = ElementAnnotationListSerializer
    openapi_overrides = {
        'operationId': 'RetrieveElementAnnotationList',
        'description': 'Retrive an IIIF annotation list for transcriptions on elements with zones',
        'security': [],
        'tags': ['iiif'],
    }

    def get_queryset(self):
        return Element.objects.filter(
            corpus__in=Corpus.objects.readable(self.request.user),
            type__folder=False,
            type__hidden=False,
            zone__isnull=False,
        )

    @method_decorator(cache_page(3600))
    def get(self, *args, **kwargs):
        return super().get(*args, **kwargs)


class TranscriptionSearchAnnotationList(SearchAPIMixin, RetrieveAPIView):
    """
    Search for transcriptions inside an element and get a IIIF annotation list
    """

    serializer_class = TranscriptionSearchAnnotationListSerializer
    query_serializer_class = IIIFSearchQuerySerializer
    openapi_overrides = {
        'operationId': 'SearchTranscriptionsAnnotationList',
        'description': 'Retrieve an IIIF Search API annotation list for transcriptions on a folder element',
        'security': [],
        'tags': ['iiif'],
    }
    elt = None

    def get_element(self):
        if not self.elt:
            self.elt = Element.objects.get(id=self.kwargs['pk'])
            if self.elt.corpus not in Corpus.objects.readable(self.request.user):
                raise PermissionDenied
        return self.elt

    def get_object(self):
        return self.get_queryset()

    def get_search(self, query=None, min_score=0.0, **kwargs):
        return ESTranscription.search() \
            .filter('match', corpus=str(self.get_element().corpus_id)) \
            .filter('match', type=TranscriptionType.Word.value) \
            .filter(
                'terms',
                element=list(
                    Element.objects
                           .filter(
                               paths__path__last=self.get_element().id,
                               type__folder=False,
                               type__hidden=False,
                               zone__isnull=False,
                           ).values_list('id', flat=True)
                )) \
            .filter('range', score={'gte': min_score}) \
            .query('match', text=query)

    def post_process(self, *args):
        return search_transcriptions_filter_post(*args, element_id=self.get_element().id)
