from math import ceil
from channels.consumer import SyncConsumer
from django.db.models import Q
from django.db.models.deletion import Collector
from django.db.models.signals import pre_delete
from arkindex.project.tools import disconnect_signal
from arkindex.project.triggers import reindex_start
from arkindex.dataimport.models import Event, Revision
from arkindex.documents.indexer import Indexer
from arkindex.documents.models import (
    Corpus, Element, ElementPath, MetaData, Transcription, TranscriptionEntity, Entity,
    EntityLink, Classification, RegionElement, Region
)
from arkindex.documents.signals import pre_delete_handler
import logging

logger = logging.getLogger(__name__)


def _delete_queryset(queryset, batch_size=1000):
    """
    Helper to delete large querysets with as little SQL queries and memory footprint as possible.
    """
    count = queryset.count()
    logger.info('Deleting {} {}'.format(count, queryset.model.__name__))

    if not count:
        return

    if Collector(using=queryset.db).can_fast_delete(queryset.all()):
        # If a single DELETE statement can be used,
        # bypass both the batched deletion and Django's related objects checks.
        logger.debug('Using single-query deletion')
        queryset._raw_delete(using=queryset.db)
        return

    if count <= batch_size:
        # If there is a single batch, just delete.
        queryset.delete()
        return

    for i in range(ceil(count / batch_size)):
        logger.debug('Deleting batch {}'.format(i + 1))
        # Deleting a slice is not allowed;
        # we use a sliced subquery instead and still delete in a single query.
        # DELETE FROM … WHERE id IN (SELECT id FROM … LIMIT [batch_size])
        ids = queryset[:batch_size].values('id')
        queryset.model.objects.filter(id__in=ids).delete()


class ReindexConsumer(SyncConsumer):

    def reindex_start(self, message):
        corpus_id, element_id, transcriptions, elements, entities, drop = map(
            message.get,
            ('corpus', 'element', 'transcriptions', 'elements', 'entities', 'drop'),
            (None, None, True, True, True, False),  # Default values
        )
        indexer = Indexer()

        if drop:
            if transcriptions:
                indexer.drop_index(Transcription.es_document)
            if elements:
                indexer.drop_index(Element.es_document)
            if entities:
                indexer.drop_index(Entity.es_document)
            indexer.setup()

        if element_id or corpus_id:
            if element_id:
                # Pick this element, and all its children
                elements_queryset = list(Element.objects.get_descending(element_id))
                elements_queryset.append(Element.objects.get(id=element_id))
            else:
                # Pick all elements in the corpus
                elements_queryset = Element.objects.filter(corpus_id=corpus_id)

            transcriptions_queryset = Transcription.objects.filter(element__in=elements_queryset)
            entities_queryset = Entity.objects.filter(
                Q(metadatas__element__in=elements_queryset)
                | Q(transcriptions__element__in=elements_queryset)
            )
        else:
            transcriptions_queryset = Transcription.objects.all()
            elements_queryset = Element.objects.all()
            entities_queryset = Entity.objects.all()

        if transcriptions:
            indexer.run_index(transcriptions_queryset, bulk_size=400)
        if elements:
            indexer.run_index(elements_queryset, bulk_size=100)
        if entities:
            indexer.run_index(entities_queryset, bulk_size=400)


class MLResultsConsumer(SyncConsumer):

    def ml_results_delete(self, message):
        corpus_id, element_id, batch_size = (
            message.get('corpus_id'),
            message.get('element_id'),
            message.get('batch_size', 1000),
        )
        assert corpus_id or element_id, 'Missing element or corpus IDs'

        if element_id:
            logger.info('Deleting ML results on element {}'.format(element_id))
            # This element and its direct children
            elements = Element.objects.filter(Q(id=element_id) | Q(paths__path__last=element_id))
            if not corpus_id:
                # The corpus ID is still used in some deletions; deduce it from the element.
                corpus_id = Element.objects.get(id=element_id).corpus_id
        elif corpus_id:
            logger.info('Deleting ML results on corpus {}'.format(corpus_id))
            elements = Element.objects.filter(corpus_id=corpus_id)

        # Simple deletions for classifications and transcriptions.
        _delete_queryset(Classification.objects.filter(element__in=elements), batch_size)
        _delete_queryset(Transcription.objects.filter(element__in=elements), batch_size)

        # Removing regions directly is not possible in an optimized way as they exist in millions
        # and any query with a JOIN, EXISTS, IN simply fails, even on dev servers.
        # The regions creation APIs just use get_or_create, so we can simply remove the links
        # between regions and elements, and ignore regions without links.
        _delete_queryset(RegionElement.objects.filter(element__in=elements), batch_size)

        # Entity deletion is complex: they can be linked to different elements both on transcriptions and metadata.
        # Metadata are not considered ML results so we need to keep them: update them to unlink entities.
        logger.info('Updating element metadata')
        MetaData.objects.filter(element__in=elements).update(entity_id=None)

        # We removed transcriptions earlier, which implies removing the links with entities.
        # All is left is to remove 'lonely' entities.
        # Note: __isnull's implementation will fetch all element IDs into a list before deleting—use batches!
        _delete_queryset(Entity.objects.filter(
            corpus_id=corpus_id,
            metadatas__isnull=True,
            transcriptions__isnull=True,
        ), batch_size)

        # Trigger a reindexation to clear up deleted results from search indexes
        reindex_start(
            element=element_id,
            corpus=corpus_id,
            transcriptions=True,
            elements=True,
            entities=True,
        )


class CorpusConsumer(SyncConsumer):

    def corpus_delete(self, message):
        corpus_id, batch_size = (
            message.get('id'),
            message.get('batch_size', 1000),
        )

        corpus = Corpus.objects.get(id=corpus_id)

        logger.info('Deleting {!r}'.format(corpus))

        # Delete all related objects, bypassing RESTRICTs deletion rules
        # and Django's way of loading everything into memory before deleting.
        querysets = [
            Event.objects.filter(element__in=corpus.elements.all()),
            corpus.imports.all(),
            Revision.objects.filter(repo__corpus_id=corpus_id),
            corpus.repos.all(),
            corpus.files.all(),
            EntityLink.objects.filter(role__corpus_id=corpus_id),
            TranscriptionEntity.objects.filter(
                Q(entity__in=corpus.entities.all())
                | Q(transcription__in=Transcription.objects.filter(element__in=corpus.elements.all()))
            ),
            corpus.entities.all(),
            corpus.roles.all(),
            corpus.allowed_metadatas.all(),
            MetaData.objects.filter(element__in=corpus.elements.all()),
            RegionElement.objects.filter(element__in=corpus.elements.all()),
            Region.objects.filter(ml_class__in=corpus.ml_classes.all()),
            Classification.objects.filter(element__in=corpus.elements.all()),
            corpus.ml_classes.all(),
            Transcription.objects.filter(element__in=corpus.elements.all()),
            ElementPath.objects.filter(element__in=corpus.elements.all()),
            corpus.elements.all(),
            corpus.types.all(),
        ]

        with disconnect_signal(pre_delete, sender=Element, receiver=pre_delete_handler):
            for queryset in querysets:
                _delete_queryset(queryset, batch_size)

        corpus.delete()
        logger.info('Deleted corpus {}'.format(corpus_id))
