from django.db import models
from itertools import groupby, chain
import uuid


class ElementManager(models.Manager):
    """Model manager for elements"""

    def get_ascending_paths(self, child_id, **filters):
        """
        Get all parent paths for a specific element ID.
        """
        # TODO: it should be possible to do in a single query through Django's filtered relations
        from arkindex.documents.models import ElementPath
        paths = ElementPath.objects.filter(element_id=child_id).values_list('path', flat=True)
        parents = {
            parent.id: parent
            for parent in self.filter(id__in=chain(*paths), **filters)
        }
        return [
            filter(None, [
                parents.get(parent_id)
                for parent_id in path
            ])
            for path in paths
        ]

    def get_ascending(self, child_id, **filters):
        """
        Get all parent elements for a specific element ID.
        > All the elements from paths
        """
        return list(chain(*self.get_ascending_paths(child_id, **filters)))

    def get_ascendings_paths(self, *children_ids, **filters):
        """
        Get all ascending paths for some elements IDs.
        """
        # Loads paths and group them by element ids
        from arkindex.documents.models import ElementPath
        paths = ElementPath.objects.filter(element_id__in=children_ids).order_by('element_id')
        tree = {
            elt_id: [p.path for p in elt_paths]
            for elt_id, elt_paths in groupby(paths, lambda e: e.element_id)
        }

        # Load parents elements in bulk
        parents = {
            parent.id: parent
            for parent in self.filter(id__in=chain(*[p.path for p in paths]), **filters)
        }
        return {
            elt_id: [
                [
                    parents[parent_id]
                    for parent_id in path
                ]
                for path in paths
            ]
            for elt_id, paths in tree.items()
        }

    def get_ascendings(self, *children_ids, **filters):
        """
        Get all parent elements for some element IDs.
        """
        # Just assemble the paths together
        return {
            elt_id: chain(*paths)
            for elt_id, paths in self.get_ascendings_paths(*children_ids, **filters).items()
        }

    def get_descending(self, parent_id, **filters):
        """
        Get all child elements for a specific element ID.
        > All the elements where parent is in the path
        """
        return self.filter(paths__path__contains=[parent_id], **filters).order_by('paths__ordering').distinct()

    def get_descendings(self, parents_ids, prefetch=(), **filters):
        """
        Get all child elements for some element IDs.
        """
        assert all(isinstance(x, uuid.UUID) for x in parents_ids), 'UUIDs are required'

        # Loads paths and group them by element ids
        # The map to str is needed to avoid a cast error in PG
        from arkindex.documents.models import ElementPath
        paths = ElementPath.objects.filter(path__overlap=map(str, parents_ids)).order_by('element_id', 'ordering')
        tree = {
            parent_id: [
                p.element_id
                for p in paths
                if parent_id in p.path
            ]
            for parent_id in parents_ids
        }

        # Load children elements in bulk
        children = self.filter(id__in=[p.element_id for p in paths], **filters)
        if prefetch:
            children = children.prefetch_related(*prefetch)
        return {
            parent_id: [
                child
                for child in children
                if child.id in tree.get(parent_id)
            ]
            for parent_id in parents_ids
        }

    def get_related(self, id, **filters):
        """
        Get all parent and child elements for a specific element ID.
        Warning : this may be very slow, don't forget to filter.
        """
        asc, desc = self.get_ascending(id, **filters), self.get_descending(id, **filters)
        return list(asc) + list(desc)

    def get_neighbors(self, element, n=1):
        """
        Get n neighbors: current, previous and next elements
        """
        neighbors = []

        from arkindex.documents.models import ElementPath
        element_path = ElementPath.objects.filter(element_id=element.id)
        for p in element_path:
            min_pos = max(p.ordering - n, 0)
            neighbors += list(ElementPath.objects.filter(
                path__contains=p.path,
                element__type=element.type
            ).order_by('ordering')[min_pos:p.ordering + n + 1])
        return neighbors


class CorpusManager(models.Manager):
    '''
    Add ACL functions to corpus listing
    '''

    def readable(self, user):
        qs = super().get_queryset().order_by('name')

        # An anonymous has access to public corpus only
        if user.is_anonymous:
            return qs.filter(public=True)

        # Admins and internal users have access to every corpus
        if user.is_admin or user.is_internal:
            return qs.all()

        # An authenticated user has access to:
        # * public corpus
        # * private with ACL
        qs = qs.prefetch_related('corpus_right')
        qs = qs.filter(models.Q(public=True) | models.Q(corpus_right__user=user)).distinct()
        return qs

    def writable(self, user):
        # An anonymous user cannot write anything
        if user.is_anonymous:
            return super().none()

        qs = super().get_queryset().order_by('name')

        # Admins and internal users have access to every corpus
        if user.is_admin or user.is_internal:
            return qs.all()

        # Authenticated users can write only on corpora with ACL
        return qs.filter(corpus_right__user=user, corpus_right__can_write=True).distinct()

    def admin(self, user):
        # An anonymous user cannot manage anything
        if user.is_anonymous:
            return super().none()

        qs = super().get_queryset().order_by('name')

        # Admins and internal users have access to every corpus
        if user.is_admin or user.is_internal:
            return qs.all()

        # Authenticated users can manage only corpora with ACL
        return qs.filter(corpus_right__user=user, corpus_right__can_admin=True).distinct()
