from rest_framework import serializers
from arkindex_common.ml_tool import MLToolType
from arkindex.project.serializer_fields import EnumField, MLToolField
from arkindex.documents.models import \
    Element, Corpus, Entity, EntityLink, EntityRole, Transcription, TranscriptionEntity
from arkindex_common.enums import EntityType
from arkindex.documents.serializers.light import \
    ElementLightSerializer, EntityLightSerializer, CorpusLightSerializer
from arkindex.documents.serializers.elements import MetaDataSerializer
from arkindex.documents.indexer import Indexer


class EntityRoleSerializer(serializers.ModelSerializer):
    """
    Serialize a role between two types of entity
    """
    parent_type = EnumField(EntityType)
    child_type = EnumField(EntityType)

    class Meta:
        model = EntityRole
        fields = (
            'id',
            'parent_name',
            'child_name',
            'parent_type',
            'child_type'
        )

    def validate(self, data):
        assert 'corpus' not in data
        assert self.context.get('request') is not None
        corpus_id = self.context['request'].parser_context['kwargs']['pk']
        corpus = Corpus.objects.writable(self.context['request'].user).filter(id=corpus_id).first()
        if corpus is None:
            raise serializers.ValidationError({
                'corpus': ['You do not have write access to this corpus'],
                'id': corpus_id,
            })
        data['corpus'] = corpus
        return data


class EntityLinkSerializer(serializers.ModelSerializer):
    """
    Serialize an entity link with its child, parent and role
    """
    parent = EntityLightSerializer()
    child = EntityLightSerializer()
    role = EntityRoleSerializer()

    class Meta:
        model = EntityLink
        fields = (
            'id',
            'parent',
            'child',
            'role'
        )


class EntitySerializer(EntityLightSerializer):
    """
    Serialize an entity with its metadata
    """
    corpus = CorpusLightSerializer()
    children = EntityLinkSerializer(many=True)
    parents = EntityLinkSerializer(many=True)
    metas = serializers.HStoreField(child=serializers.CharField(), allow_null=True, required=False)

    class Meta:
        model = Entity
        fields = EntityLightSerializer.Meta.fields + (
            'corpus',
            'metas',
            'children',
            'parents'
        )

    def validate(self, data):
        moderator = self.context['request'].user if data['validated'] else None
        return {
            'validated': data['validated'],
            'moderator': moderator
        }


class EntityCreateSerializer(EntityLightSerializer):
    """
    Serialize an entity with a possible parents and children
    """
    corpus = serializers.PrimaryKeyRelatedField(queryset=Corpus.objects.none())
    metas = serializers.HStoreField(child=serializers.CharField(), required=False)
    children = EntityLinkSerializer(many=True, read_only=True)
    parents = EntityLinkSerializer(many=True, read_only=True)
    ner = MLToolField(MLToolType.NER)

    class Meta:
        model = Entity
        fields = (
            'id',
            'name',
            'type',
            'metas',
            'validated',
            'corpus',
            'parents',
            'children',
            'ner'
        )
        read_only_fields = ('id', )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        self.fields['corpus'].queryset = corpora

    def save(self, **kwargs):
        instance = super().save(**kwargs)
        Indexer().run_index(
            Entity.objects.filter(id=instance.id),
        )
        return instance


class EntityLinkCreateSerializer(EntityLinkSerializer):
    """
    Serialize an entity with a possible parents and children
    """
    parent = serializers.PrimaryKeyRelatedField(queryset=Entity.objects.none())
    child = serializers.PrimaryKeyRelatedField(queryset=Entity.objects.none())
    role = serializers.PrimaryKeyRelatedField(queryset=Entity.objects.none())

    class Meta:
        model = EntityLink
        fields = EntityLinkSerializer.Meta.fields

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Do not raise Error in order to create OpenAPI schema
            return
        corpora = Corpus.objects.writable(self.context['request'].user)
        entities = Entity.objects.all().filter(corpus__in=corpora)
        roles = EntityRole.objects.all().filter(corpus__in=corpora)
        self.fields['parent'].queryset = entities
        self.fields['child'].queryset = entities
        self.fields['role'].queryset = roles

    def validate(self, data):
        data = super().validate(data)
        link = EntityLink(**data)
        link.full_clean()
        return data


class TranscriptionEntitySerializer(serializers.ModelSerializer):
    """
    Serialise the link between an entity and a transcription
    Transcription element must be passed in serializer context
    """
    offset = serializers.IntegerField(min_value=0)
    length = serializers.IntegerField(min_value=0)

    class Meta:
        model = TranscriptionEntity
        fields = ('entity', 'offset', 'length')

    def validate(self, data):
        data = super().validate(data)
        data['transcription'] = self.context['transcription']
        if not data['transcription']:
            raise serializers.ValidationError({
                'transcription': ['invalid UUID or Corpus write-access is forbidden']
            })
        return data


class TranscriptionEntityDetailsSerializer(TranscriptionEntitySerializer):
    """
    Serialize the link between a transcription and a detailled entity
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.fields['entity'] = EntityLightSerializer()


class ElementEntitiesSerializer(ElementLightSerializer):
    """
    Serialize an element with its entities
    Entities are split into two types, those extracted
    from metadata and those extracted from element transcriptions
    """
    metadata = MetaDataSerializer(many=True)
    transcriptions = TranscriptionEntityDetailsSerializer(many=True)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):
            # Keep original fields so OpenAPI can build its schemas
            return
        self.fields['transcriptions'] = serializers.SerializerMethodField()
        self.fields['metadata'] = serializers.SerializerMethodField()

    def get_transcriptions(self, obj):
        elt_transcriptions = Transcription.objects.filter(element_id=obj.id)
        transcription_entities = TranscriptionEntity.objects \
            .filter(transcription__in=elt_transcriptions) \
            .prefetch_related('entity') \
            .order_by('offset')
        return TranscriptionEntityDetailsSerializer(transcription_entities, many=True).data

    def get_metadata(self, obj):
        metadata = obj.metadatas \
            .filter(entity__isnull=False) \
            .prefetch_related('revision__repo', 'entity')
        return MetaDataSerializer(metadata, many=True).data

    class Meta:
        model = Element
        fields = ElementLightSerializer.Meta.fields + ('transcriptions', 'metadata')
