from rest_framework import serializers
from rest_framework.exceptions import ValidationError, APIException
from django.db.models import Max
from arkindex.documents.models import (
    Element, ElementType, Corpus, Entity, Region,
    MetaData, AllowedMetaData
)
from arkindex_common.enums import EntityType, MetaType
from arkindex.documents.dates import DateType
from arkindex.documents.serializers.ml import DataSourceSerializer, MLClassSerializer
from arkindex.images.serializers import ZoneSerializer
from arkindex.dataimport.serializers.git import RevisionSerializer
from arkindex.project.serializer_fields import EnumField


class InterpretedDateSerializer(serializers.Serializer):
    """
    Serialize a list of interpreted dates linked to a metadata
    """
    type = EnumField(DateType)
    year = serializers.IntegerField(min_value=0)
    month = serializers.IntegerField(min_value=1, max_value=12, allow_null=True)
    day = serializers.IntegerField(min_value=1, max_value=31, allow_null=True)

    class Meta:
        fields = (
            'year',
            'month',
            'day',
            'type',
        )


class ElementLightSerializer(serializers.ModelSerializer):
    """
    Serialises a Element
    """
    type = serializers.CharField(source='type.slug')

    class Meta:
        model = Element
        fields = (
            'id',
            'type',
            'name',
        )


class ElementTypeSerializer(serializers.ModelSerializer):

    class Meta:
        model = ElementType
        fields = (
            'slug',
            'display_name',
            'folder',
            'hidden',
            'default_view',
        )


class CorpusLightSerializer(serializers.ModelSerializer):
    """
    Serializes a Corpus
    """

    class Meta:
        model = Corpus
        fields = (
            'id',
            'name',
            'public',
        )


class EntityLightSerializer(serializers.ModelSerializer):
    """
    Serializes an entity
    """
    type = EnumField(EntityType)
    dates = InterpretedDateSerializer(many=True, source='get_dates', required=False, allow_null=True)
    metas = serializers.HStoreField(child=serializers.CharField(), required=False, allow_null=True)
    source = DataSourceSerializer()

    class Meta:
        model = Entity
        fields = (
            'id',
            'name',
            'type',
            'metas',
            'validated',
            'dates',
            'source'
        )


class RegionLightSerializer(serializers.ModelSerializer):
    """
    Serialize a region without its related elements
    """
    ml_class = MLClassSerializer()
    source = DataSourceSerializer()
    zone = ZoneSerializer()

    class Meta:
        model = Region
        fields = (
            'id',
            'source',
            'ml_class',
            'confidence',
            'zone',
        )


class DuplicatedMetadata(APIException):
    status_code = 200
    default_detail = 'Metadata exists.'


class MetaDataLightSerializer(serializers.ModelSerializer):
    """
    Serialize a metadata without its entity
    """
    type = EnumField(MetaType)
    revision = RevisionSerializer(read_only=True)
    dates = InterpretedDateSerializer(many=True, source='get_dates', read_only=True)

    def check_allowed(self, corpus=None, type=None, name=None, instance=None, **kwargs):
        """
        Assert metadata type and value are allowed in this corpus
        """
        message = 'Metadata with this (type, name) are forbidden for this corpus'
        if instance:
            type = instance.type
            name = instance.name
            message = 'Edition of metadata with this (type, name) are disallowed for this corpus'
        if not AllowedMetaData.objects.filter(corpus=corpus, type=type, name=name):
            raise ValidationError({
                '__all__': [message]
            })

    def create(self, validated_data):
        user = self.context['request'].user
        element = self.context['element']
        if not (user.is_admin or user.is_internal):
            self.check_allowed(corpus=element.corpus, **validated_data)
        metadata = element.metadatas.filter(name=validated_data['name'])
        if metadata.exists():
            if metadata.filter(value=validated_data['value'], type=validated_data['type']).exists():
                # A metdata with same type, name and value exists, raises a 200_OK
                raise DuplicatedMetadata
            validated_data['index'] = metadata.aggregate(Max('index'))['index__max'] + 1
        metadata = element.metadatas.create(**validated_data)
        return metadata

    def update(self, instance, validated_data):
        user = self.context['request'].user
        element = instance.element
        if not (user.is_admin or user.is_internal):
            # Assert actual instance is part of AllowedMetaData
            self.check_allowed(corpus=element.corpus, instance=instance)
            # Assert update is part of AllowedMetaData
            self.check_allowed(corpus=element.corpus, **validated_data)
        new_name = validated_data.get('name') or instance.name
        metadata = element.metadatas \
            .filter(name=new_name) \
            .exclude(id=instance.id)
        if metadata.exists():
            validated_data['index'] = metadata.aggregate(Max('index'))['index__max'] + 1
        return super().update(instance, validated_data)

    def delete(self, instance):
        user = self.context['request'].user
        element = instance.element
        if not (user.is_admin or user.is_internal):
            self.check_allowed(corpus=element.corpus, instance=instance)
        instance.delete()

    class Meta:
        model = MetaData
        read_only_fields = ('id', )
        fields = (
            'id',
            'type',
            'name',
            'value',
            'revision',
            'dates',
        )


class CorpusAllowedMetaDataSerializer(serializers.ModelSerializer):
    """
    Serialize allowed (type, name) couples for metadata creation in a corpus
    """
    type = EnumField(MetaType)

    class Meta:
        model = AllowedMetaData
        fields = ('type', 'name', )
        read_only_fields = ('type', 'name', )
