from enum import Enum
from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from arkindex_common.ml_tool import MLToolType
from arkindex_common.enums import TranscriptionType
from arkindex.project.serializer_fields import EnumField, MLToolField, PolygonField
from arkindex.documents.models import (
    Corpus, Element, Transcription, DataSource, MLClass, Classification, ClassificationState
)
from arkindex.images.serializers import ZoneSerializer


class ClassificationMode(Enum):
    """
    Classification mode
    """
    Validate = "validate"
    Create = "create"


class DataSourceSerializer(serializers.ModelSerializer):
    """
    Serialize a data source for transcriptions and classifications
    """

    type = EnumField(MLToolType)

    class Meta:
        model = DataSource
        fields = (
            'id',
            'type',
            'slug',
            'name',
            'revision',
            'internal',
        )


class DataSourceStatsSerializer(DataSourceSerializer):
    """
    A data source, but including their result counts
    """
    classifications_count = serializers.IntegerField(default=0)
    transcriptions_count = serializers.IntegerField(default=0)
    regions_count = serializers.IntegerField(default=0)
    entities_count = serializers.IntegerField(default=0)

    class Meta(DataSourceSerializer.Meta):
        fields = DataSourceSerializer.Meta.fields + (
            'classifications_count',
            'transcriptions_count',
            'regions_count',
            'entities_count',
        )


class MLClassSerializer(serializers.ModelSerializer):
    """
    Serializer for MLClass instances
    Used as a nested endpoint below corpus, so exclude corpus from payload
    """

    class Meta:
        model = MLClass
        exclude = ('corpus',)


class CountMLClassSerializer(MLClassSerializer):
    nb_best = serializers.IntegerField()


class ClassificationSerializer(serializers.ModelSerializer):
    """
    Serialize a classification on an Element
    """

    source = DataSourceSerializer()
    ml_class = MLClassSerializer()
    state = EnumField(ClassificationState)

    class Meta:
        model = Classification
        read_only_fields = ('id', 'confidence', 'high_confidence')
        fields = (
            'id',
            'source',
            'ml_class',
            'state',
            'confidence',
            'high_confidence'
        )


class ClassificationCreateSerializer(serializers.ModelSerializer):
    """
    Specific serializer for the manual classification creation
    """
    class Meta:
        model = Classification
        fields = (
            'element',
            'ml_class',
        )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return
        self.fields['element'].queryset = Element.objects.filter(
            corpus__in=Corpus.objects.writable(self.context['request'].user))


class ClassificationsSelectionSerializer(serializers.ModelSerializer):
    """
    Serializer for the manual classification creation on the selection
    """
    ml_class = serializers.UUIDField(required=False)
    corpus_id = serializers.UUIDField()
    mode = EnumField(ClassificationMode)

    class Meta:
        model = Classification
        fields = (
            'corpus_id',
            'mode',
            'ml_class'
        )

    def validate(self, data):
        data = super().validate(data)
        corpus = Corpus.objects.get(id=data['corpus_id'])

        if data['mode'] == ClassificationMode.Create:
            ml_class = MLClass.objects.filter(id=data['ml_class']).first()
            if ml_class is None:
                raise ValidationError({'ml_class': 'Ml class {} not found'.format(data['ml_class'])})
            if ml_class.corpus.id != corpus.id:
                raise ValidationError('Ml class {} does not belong to the corpus {}'.format(
                    ml_class,
                    corpus
                ))
        return data


class TranscriptionSerializer(serializers.ModelSerializer):
    """
    Serialises a Transcription
    issued from a search
    """

    type = EnumField(TranscriptionType)
    zone = ZoneSerializer()
    source = DataSourceSerializer()

    class Meta:
        model = Transcription
        fields = (
            'id',
            'type',
            'text',
            'score',
            'zone',
            'source',
        )


class TranscriptionCreateSerializer(serializers.Serializer):
    """
    Allows for insertion of new transcriptions and zones
    """
    element = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.filter(zone__isnull=False),
    )
    source = serializers.PrimaryKeyRelatedField(
        queryset=DataSource.objects.filter(type=MLToolType.Recognizer),
    )
    polygon = PolygonField()
    text = serializers.CharField()
    score = serializers.FloatField(min_value=0, max_value=1)
    type = EnumField(TranscriptionType)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return
        self.fields['element'].queryset = Element.objects.filter(
            zone__isnull=False,
            corpus__in=Corpus.objects.writable(self.context['request'].user),
        )


class TranscriptionBulkSerializer(serializers.Serializer):
    """
    Allows for insertion of new transcriptions and zones
    in Bulk (used by serializer below)
    Note: no element !
    """
    polygon = PolygonField()
    text = serializers.CharField()
    score = serializers.FloatField(min_value=0, max_value=1)
    type = EnumField(TranscriptionType)


class TranscriptionsSerializer(serializers.Serializer):
    """
    Allows for insertion of new transcriptions and zones
    in Bulk (uses serializer above) on a common parent
    """
    parent = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.filter(zone__isnull=False),
    )
    recognizer = MLToolField(MLToolType.Recognizer)
    transcriptions = TranscriptionBulkSerializer(many=True, allow_empty=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return
        self.fields['parent'].queryset = Element.objects.filter(
            zone__isnull=False,
            corpus__in=Corpus.objects.writable(self.context['request'].user),
        )


class ClassificationBulkSerializer(serializers.Serializer):
    """
    Single classification serializer for bulk insertion
    Cannot use ModelSerializer as they become read-only when nested
    """
    class_name = serializers.CharField(source='ml_class')
    confidence = serializers.FloatField(min_value=0, max_value=1)
    high_confidence = serializers.BooleanField(default=False)


class ClassificationsSerializer(serializers.Serializer):
    """
    Insert N classifications on a single element from a single ML tool
    """
    parent = serializers.PrimaryKeyRelatedField(
        queryset=Element.objects.all(),
    )
    classifier = MLToolField(MLToolType.Classifier)
    classifications = ClassificationBulkSerializer(many=True, allow_empty=False)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not self.context.get('request'):  # May be None when generating an OpenAPI schema or using from a REPL
            return
        self.fields['parent'].queryset = Element.objects.filter(
            corpus__in=Corpus.objects.writable(self.context['request'].user),
        )

    def validate(self, data):
        data = super().validate(data)
        corpus_id = data['parent'].corpus.id
        for classification in data['classifications']:
            # Retrieve or create required ml_class
            classification['ml_class'], _ = MLClass.objects.get_or_create(
                name=classification['ml_class'],
                corpus_id=corpus_id
            )
        return data
