from itertools import groupby
from arkindex_common.tei import TeiElement, TeiParser as BaseTeiParser
from arkindex_common.enums import MetaType
from arkindex.project.tools import find_closest
from arkindex.project.triggers import reindex_start
from arkindex.documents.models import Element, Entity, DataSource, MLToolType
import logging


logger = logging.getLogger(__name__)


class TeiParser(BaseTeiParser):

    def __init__(self, path, corpus, revision, folder_type, element_type):
        super().__init__(path)
        assert folder_type.corpus == corpus and folder_type.folder
        assert element_type.corpus == corpus
        self.db_corpus = corpus
        self.revision = revision
        self.folder_type = folder_type
        self.element_type = element_type

    def match_database(self):
        '''
        Link TEI elements with DB elements
        Outputs a list of tuples (db_element, tei_element)
        '''
        folders = self.db_corpus.elements.filter(type=self.folder_type)
        assert folders.count() > 0, 'No folders available'

        # Match folders
        out = []
        for tei in self.corpus.tei:
            if not tei.witness:
                logger.warning('No witness in {}'.format(str(tei)))
                continue

            tei_name = tei.witness.id or tei.witness.repository_id
            folder = find_closest(tei_name, folders)
            if not folder:
                logger.warning('No matching folder for {}'.format(tei))
                continue

            out.append((folder, tei))
            logger.info('Matched {} with {}'.format(folder, tei))

            # Load folder elements
            folder_elements = Element.objects.get_descending(folder.id, type=self.element_type)
            if not folder_elements.exists():
                logger.warning('No elements in DB for {}'.format(folder))
                continue

            # Match elements
            for text in tei.texts:
                if text.witness is None:
                    logger.warning('No witness on text, skipping.')
                    continue

                element = Element.objects.filter(
                    id__in=folder_elements,
                    type=self.element_type,
                    metadatas__name='number',
                    metadatas__value=text.witness.id,
                ).first()
                if element:
                    out.append((element, text))
                    logger.info('Matched {} with {}'.format(element, text))
                else:
                    logger.warning('No match for {}'.format(text))

        return out

    def save(self, tei_elt, db_elt):
        """
        Save or update all metadatas contained in TEI element to corresponding database element
        Return a list of saved Metadatas
        """
        assert isinstance(tei_elt, TeiElement)
        assert isinstance(db_elt, Element)

        # Filter to remove empty values
        parsed_metadatas = list(filter(lambda meta: meta.value, tei_elt.build_metadata()))

        # Group by name
        def grouper(metadata):
            return metadata.name

        parsed_metadatas = groupby(sorted(parsed_metadatas, key=grouper), key=grouper)
        db_metadatas = []

        # Create/update
        for name, metadatas in parsed_metadatas:
            for index, xmlmeta in enumerate(metadatas):
                # Handle entities
                entity = None
                if xmlmeta.entity_type is not None:
                    source, _ = DataSource.objects.get_or_create(
                        type=MLToolType.NER,
                        slug='TEI',
                        internal=False
                    )
                    entity, _ = Entity.objects.get_or_create(
                        corpus=db_elt.corpus,
                        name=xmlmeta.value,
                        type=xmlmeta.entity_type,
                        source=source
                    )

                db_meta, created = db_elt.metadatas.get_or_create(
                    name=name,
                    index=index,
                    defaults={
                        'type': xmlmeta.type,
                        'value': xmlmeta.value,
                        'revision': self.revision,
                        'entity': entity,
                    },
                )
                db_metadatas.append(db_meta)
                if created:
                    continue
                if (db_meta.type, db_meta.value) == (xmlmeta.type, xmlmeta.value) \
                        and db_meta.revision \
                        and (not db_meta.entity or db_meta.entity.name == xmlmeta.value):
                    # Nothing to update and revision is set
                    continue
                db_meta.type, db_meta.value, db_meta.entity, db_meta.revision = \
                    xmlmeta.type, xmlmeta.value, entity, self.revision
                db_meta.save()

        # Remove deleted metadatas
        db_elt.metadatas.exclude(
            id__in=[dbm.id for dbm in db_metadatas],
        ).delete()
        return db_metadatas

    def run(self):
        matches = self.match_database()
        logger.info('Found {} metadatas matching DB'.format(len(matches)))
        db_metadatas = []
        for db_elt, tei_elt in matches:
            db_metadatas += self.save(tei_elt, db_elt)
        self.index(db_metadatas)

    def index(self, db_metadatas):
        """
        Create or Update ElasticSearch index for indexed elements
        """
        md_ids = (md.id for md in db_metadatas if md.type == MetaType.Date)
        elements = Element.objects.filter(metadatas__id__in=md_ids, type__folder=False, type__hidden=False)
        if elements.exists():
            reindex_start(corpus=self.db_corpus, elements=True)
