from unittest.mock import patch
from asyncmock import AsyncMock
from django.urls import reverse
from rest_framework import status
from arkindex.project.tests import FixtureTestCase
from arkindex.documents.models import Corpus


class TestAdminAPI(FixtureTestCase):

    def test_reindex_requires_login(self):
        response = self.client.post(reverse('api:reindex-start'), {})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_reindex_requires_admin(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse('api:reindex-start'), {})
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_reindex_requires_index_types(self):
        self.client.force_login(self.superuser)
        response = self.client.post(reverse('api:reindex-start'), {
            'transcriptions': False,
            'elements': False,
            'entities': False,
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['At least one index type is required.'],
        })

    def test_reindex_drop_only_all(self):
        self.client.force_login(self.superuser)
        response = self.client.post(reverse('api:reindex-start'), {
            'corpus': str(self.corpus.id),
            'drop': True,
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['`drop` can only be used when reindexing everything.'],
        })

    def test_reindex_element_in_corpus(self):
        corpus2 = Corpus.objects.create(name='some corpus')
        self.client.force_login(self.superuser)
        response = self.client.post(reverse('api:reindex-start'), {
            'corpus': str(corpus2.id),
            'element': str(self.corpus.elements.first().id),
        })
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'non_field_errors': ['The selected element is not in the selected corpus.'],
        })

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_reindex(self, get_layer_mock):
        get_layer_mock.return_value.send = AsyncMock()

        self.client.force_login(self.superuser)
        response = self.client.post(reverse('api:reindex-start'), {})
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        get_layer_mock.return_value.send.assert_called_once_with(
            'reindex',
            {
                'element': None,
                'corpus': None,
                'transcriptions': True,
                'elements': True,
                'entities': True,
                'drop': False,
                'type': 'reindex.start',
            },
        )
