from collections import namedtuple
from django.db import connection
from arkindex_common.enums import TranscriptionType
from arkindex.project.polygon import Polygon
from arkindex.documents.models import Element
from arkindex.images.models import Image, Zone
import csv
import io
import os
import re
import gzip
import logging
import uuid

REGEX_INDEX = re.compile(
    br'^(?:line_(\d+) )?(.+) \d+ ([\de\-\.]+) (\d+) (\d+) (\d+) (\d+)')

logger = logging.getLogger(__name__)

# Link a transcription data with a Polygon
# This is hashable (box is hashable)
TrPolygon = namedtuple('TrPolygon', 'image_id, element_id, source_id, type, polygon, text, score')


def extract_transcriptions(image, index_path):
    """
    Extract transcription data from index files
    One gzipped index file per image
    Format: #line word n/a score-[0,1] x1 y1 x2 y2
    """
    assert isinstance(image, Image)
    assert os.path.exists(index_path), \
        'Missing index {}'.format(index_path)
    assert index_path.endswith('.idx.gz')

    # Parse lines content
    lines = []
    with gzip.open(index_path) as f:
        for line in f.readlines():
            if line.decode('utf-8').startswith('#'):
                continue
            index = REGEX_INDEX.match(line)
            if index is None:
                logger.warning('Index parsing failed : {}'.format(line))
                continue

            # Build zone
            data = index.groups()
            x, y = int(data[3]), int(data[4])
            lines.append({
                'x': x,
                'y': y,
                'width': int(data[5]) - x,
                'height': int(data[6]) - y,
                'text': data[1].decode('utf-8'),
                'line': int(data[0] or 0),
                'score': float(data[2]),
            })

    logger.debug('Parsed {} lines'.format(len(lines)))
    return lines


def build_transcriptions(image, parent, source, items):
    """
    Build needed TrPolygon namedtuples from parsed index files
    """
    assert isinstance(parent, Element)

    # Build all TrPolygon from items
    required = {
        TrPolygon(
            image.id,
            parent.id,
            source.id,
            i.get('type', TranscriptionType.Word),
            Polygon.from_dict(i),
            i['text'],
            float(i['score']),
        )
        for i in items
    }

    # Build all TrPolygon from existing
    existing = {
        TrPolygon(
            tr.zone.image_id,
            tr.element_id,
            tr.source_id,
            tr.type,
            tr.zone.polygon,
            tr.text,
            tr.score,
        )
        for tr in parent.transcriptions.all().prefetch_related('zone')
    }

    # Calc needed TrPolygon to build
    return list(required.difference(existing))


def save_transcriptions(*tr_polygons, delimiter='\t', quotechar='"'):
    """
    Save TrPolygons in the database
    Uses PostgreSQL's COPY, as it is the fastest way to import, by generating CSV data in file-like objects
    """

    assert len(tr_polygons) > 0, \
        'No polygons, so no transcriptions to save.'

    def _list_to_csv(rows):
        """
        Convert a two dimensional iterable to a PostgreSQL COPY-compatible file-like object
        """
        out = io.StringIO()
        writer = csv.writer(out, dialect='unix', delimiter=delimiter, quotechar=quotechar, quoting=csv.QUOTE_MINIMAL)
        writer.writerows(rows)
        out.seek(0)
        return out

    zone_ids = {
        (image_id, polygon): zone_id
        for zone_id, image_id, polygon
        in Zone.objects
               .filter(image_id__in={tr_polygon.image_id for tr_polygon in tr_polygons})
               .values_list('id', 'image_id', 'polygon')
    }

    new_zones, transcriptions = [], []
    for tr_polygon in tr_polygons:
        key = (tr_polygon.image_id, tr_polygon.polygon)
        zone_id = zone_ids.get(key)
        if not zone_id:
            zone_id = uuid.uuid4()
            zone_ids[key] = zone_id
            new_zones.append(Zone(
                id=zone_id,
                image_id=tr_polygon.image_id,
                polygon=tr_polygon.polygon,
            ))

        transcriptions.append((
            uuid.uuid4(),
            tr_polygon.element_id,
            tr_polygon.source_id,
            zone_id,
            tr_polygon.type.value,
            tr_polygon.text,
            tr_polygon.score,
        ))

    Zone.objects.bulk_create(new_zones)

    # Convert to CSV and run COPY via psycopg2
    with connection.cursor() as cursor:
        cursor.copy_from(
            _list_to_csv(transcriptions),
            'documents_transcription',
            columns=('id', 'element_id', 'source_id', 'zone_id', 'type', 'text', 'score'),
            sep=delimiter,
        )

    return transcriptions
