from arkindex_common.enums import MetaType
from collections.abc import Sequence
from itertools import chain
from django.conf import settings
from elasticsearch_dsl import Document, InnerDoc, Mapping, Text, Float, Keyword, Nested, Date, DateRange, Search
from elasticsearch_dsl.connections import connections

connections.create_connection(hosts=settings.ELASTIC_SEARCH_HOSTS)


class ESQuerySet(Sequence):

    def __init__(self, func, search):
        assert isinstance(search, Search)
        self.func = func
        self.search = search

    def __iter__(self):
        # When requesting to iterate over the results without explicitly setting
        # a size, ES will only return a few results, messing up the IIIF search API.
        return iter(self.func(self.search[0:settings.ES_RESULTS_LIMIT].execute()))

    def __getitem__(self, value):
        return self.func(self.search[value].execute())

    def __len__(self):
        return self.search.count()


class RawDate(Date):
    """
    A Date field that just returns strings. Useful to use the ES date math
    https://www.elastic.co/guide/en/elasticsearch/reference/6.2/common-options.html#date-math
    """
    # TODO: Handle InterpretedDates?

    def _deserialize(self, data):
        if isinstance(data, str):
            if '||' in data:
                data = data.partition('||')[0]
            return data
        return super()._deserialize(data)


class RawDateRange(DateRange):
    _core_field = RawDate()


class ESTranscription(Document):
    element = Keyword()
    corpus = Keyword()
    type = Keyword()
    score = Float()
    text = Text()

    class Meta:
        mapping = Mapping('transcription')

    class Index:
        name = 'transcriptions'

    @classmethod
    def from_model(cls, instance):
        return cls(
            meta={'id': instance.id.hex},
            element=instance.element_id,
            type=instance.type.value,
            score=instance.score,
            text=instance.text,
            corpus=instance.element.corpus_id,
        )


class ESTranscriptionInnerDoc(InnerDoc):
    """
    A transcription nested inside an element document.
    """
    id = Keyword()
    type = Keyword()
    score = Float()
    text = Text()

    @classmethod
    def from_model(cls, instance):
        return cls(
            id=instance.id,
            type=instance.type.value,
            score=instance.score,
            text=instance.text,
        )


class ESElement(Document):
    type = Keyword()
    corpus = Keyword()
    # Used exclusively for sorting
    parents = Keyword()
    references = Keyword()
    transcriptions = Nested(ESTranscriptionInnerDoc)
    date_range = RawDateRange(format='yyyy||yyyy-MM||yyyy-MM-dd')

    class Meta:
        mapping = Mapping('element')

    class Index:
        name = 'elements'

    @classmethod
    def from_model(cls, instance):
        from arkindex.documents.models import Element, Transcription

        hidden_children = Element \
            .objects \
            .get_descending(instance.id) \
            .filter(type__hidden=True, zone__isnull=False)
        transcriptions = chain(
            instance.transcriptions.all(),
            *[
                Transcription.objects.filter(
                    zone__image=elt.zone.image,
                    zone__polygon__in=elt.zone.polygon
                )
                for elt in hidden_children
            ]
        )
        interpreted_dates = chain(*[md.get_dates() for md in instance.metadatas.all()])
        date_range = {
            k: val
            for date in interpreted_dates
            for k, val in date.to_es_range().items()
        }
        if not date_range:
            # Prevent inserting an infinite date range (an empty dict)
            date_range = None

        return cls(
            meta={'id': instance.id.hex},
            corpus=instance.corpus_id,
            type=instance.type.slug,
            parents=[
                element.name
                for element in Element.objects.get_ascending(instance.id)
            ],
            references=[
                md.value.lower()
                for md in instance.metadatas.filter(type=MetaType.Reference)
            ],
            transcriptions=list(map(
                ESTranscriptionInnerDoc.from_model,
                transcriptions,
            )),
            date_range=date_range,
        )


class ESEntity(Document):
    corpus = Keyword()
    type = Keyword()
    name = Text()

    class Meta:
        mapping = Mapping('entity')

    class Index:
        name = 'entities'

    @classmethod
    def from_model(cls, instance):
        return cls(
            meta={'id': instance.id.hex},
            corpus=instance.corpus_id,
            type=instance.type.value,
            name=instance.name,
        )
