from django.shortcuts import get_object_or_404
from django.core.exceptions import PermissionDenied
from arkindex.documents.models import Corpus, Right, Element
from arkindex.documents.serializers.search import SearchQuerySerializer
from arkindex.project.elastic import ESQuerySet
from arkindex.project.openapi import SearchAutoSchema


class CorpusACLMixin(object):

    def get_corpus(self, corpus_id, right=Right.Read):
        corpus = get_object_or_404(Corpus, id=corpus_id)
        if right not in corpus.get_acl_rights(self.request.user):
            raise PermissionDenied()
        return corpus

    def has_read_access(self, corpus):
        assert isinstance(corpus, Corpus)
        return Right.Read in corpus.get_acl_rights(self.request.user)

    def has_write_access(self, corpus):
        assert isinstance(corpus, Corpus)
        return Right.Write in corpus.get_acl_rights(self.request.user)

    def has_admin_access(self, corpus):
        assert isinstance(corpus, Corpus)
        return Right.Admin in corpus.get_acl_rights(self.request.user)


class NestedCorpusMixin(CorpusACLMixin):
    """
    Used for corpora nested endpoints.
    Fetch the corpus instance once, and attach it to both view and request.
    """
    def initialize_request(self, request, *args, **kwargs):
        request = super().initialize_request(request, *args, **kwargs)
        request.corpus = self.corpus = self.get_corpus(self.kwargs['pk'])
        return request


class SearchAPIMixin(CorpusACLMixin):
    query_serializer_class = SearchQuerySerializer
    schema = SearchAutoSchema()
    search = None

    def get_search(self, **query):
        return self.search

    def get_queryset(self):
        serializer = self.query_serializer_class(
            context={'request': self.request},
            data=self.request.query_params,
        )
        serializer.is_valid(raise_exception=True)
        query = serializer.validated_data
        # TODO Handle corpus field in serializer too
        if query.get('corpus_id'):
            try:
                query['corpora_ids'] = [str(self.get_corpus(query['corpus_id']).id), ]
            except Corpus.DoesNotExist:
                raise PermissionDenied
        else:
            query['corpora_ids'] = list(map(
                str,
                Corpus.objects.readable(self.request.user).values_list('id', flat=True),
            ))

        query.pop('corpus_id', None)
        search = self.get_search(**query)
        return ESQuerySet(self.post_process, search)

    def post_process(self, *args):
        return args


class SelectionMixin(CorpusACLMixin):

    def get_selection(self, corpus_id=None):
        elements = self.request.session.get('elements_selected')
        if elements is None:
            return []
        if corpus_id is None:
            return Element.objects.filter(id__in=elements)
        return Element.objects.filter(id__in=elements, corpus=corpus_id)
